

#include <stdio.h>
#include <assert.h>
#include "newthreads.h"

#ifndef GNUCC
#define inline /**/
#endif

Thread* thisThread;

static Thread* readyQhead = (Thread*)0;
static Thread* readyQtail = (Thread*)0;
static Spinlock readyQlock;


Thread dummyStartupThread;
Thread idleThread;
Thread mainThread;


char* newthreadsVersion()
{
  return "2.0";
}


static inline Thread* readyQget()
{
  Thread* ret;

  AcquireSpinlock(readyQlock);
  ret = readyQhead;
  if(ret){
	readyQhead = ret->readyQ;
	while(ret->status == Running)    ;
  }
  ReleaseSpinlock(readyQlock);
  return ret;
}


static inline void readyQput(Thread* t)
{
  t->readyQ = (Thread*)0;
  AcquireSpinlock(readyQlock);
  if(readyQhead){
	readyQtail->readyQ = t;
	readyQtail = t;
  }else{
	readyQhead = readyQtail = t;
  }
  ReleaseSpinlock(readyQlock);
}


int saveArgc;
char** saveArgv;
void initThreadPackage();
void _initThreadPackage();

#ifdef STANDALONE
/*
 * Initialize newthreads and start userMain() as
 * the first task.  We end up blocked.
 */
void main(int argc, char* argv[])
{

#ifndef PVM
  saveArgc = argc;
  saveArgv = argv;
#endif
  initCommStuff();
  initThreadPackage();
}
#else
/*
 * Initialize newthreads, return control to caller
 */
void init_newthreads(int argc, char* argv[])
{
#ifndef PVM
  saveArgc = argc;
  saveArgv = argv;
#endif
  initCommStuff();
  _initThreadPackage();
}
#endif


void _initThreadPackage()
  /* some sleight-of-hand to get ourselves into the thread package */
{
  void idleThreadFunc(void*);

  InitSpinlock(readyQlock);
  sprintf(dummyStartupThread.name, "_dummyStartupThread");
  sprintf(idleThread.name, "_idleThread");
  thisThread = &dummyStartupThread;
  threadStart(&idleThread, idleThreadFunc, (void*)0);
}

void initThreadPackage()
{
  /* some sleight-of-hand to get ourselves into the thread package */
  _initThreadPackage();
  threadBlock();
}


#ifdef IntelNX
#define NumSkipDeliveries  4
#endif
#ifdef PVM
#define NumSkipDeliveries  20
#endif
#ifdef PVM3
#define NumSkipDeliveries  20
#endif

void idleThreadFunc(void* ignoredArg)
  /* executed by the idle-thread */
{
  int numToSkip = 0;
  extern int numRecvWaiters;
#ifdef STANDALONE
  void mainThreadFunc(void*);
#endif

  threadUsesFloatingPoint(0);
#ifdef STANDALONE
  sprintf(mainThread.name, "_mainThread");
  threadStart(&mainThread, mainThreadFunc, (void*)0);
#endif
  for(;;){
    threadYield();
#ifndef OneNode
	if(numRecvWaiters > 0){
	  if( (readyQhead == (Thread*)0) || (numToSkip-- <= 0) ){
		deliverMessages();
		numToSkip = NumSkipDeliveries;
	  }
	}
#endif
  }
  assert(0);
}


#ifdef STANDALONE
void mainThreadFunc(void* ignoredArg)
{
  void userMain(int, char*[]);

  userMain(saveArgc, saveArgv);
  exit(0);
}
#endif


void threadStart(Thread* t, ThreadFunction func, void* arg)
{
  assert(t);
  assert(t->status == Defunct);
  
  t->status = Virgin;
  t->startFunc = func;
  t->startArg = arg;
  t->usesFloatingPoint = 1;

  /* Initialize the stack pointer.
   * Be sure to leave a little slop so we don't write past the end of
   * the stack when a thread's start function is called.
   */
  if(StackGrowsUp){
	t->saveSp = &(t->stack[32]);
  }else{
	t->saveSp = &(t->stack[ThreadStackSize-32]);
  }
  readyQput(t);
}


void threadBlock()
{
  Thread* nextGuy;

  nextGuy = readyQget();
  assert(nextGuy);
  threadHandoffCPU(nextGuy, Blocked);
}


void threadUnblock(Thread* t)
{
  while(t->status == Running)    ;
  readyQput(t);
}


void threadYield()
{
  Thread* nextGuy;

  nextGuy = readyQget();
  if(nextGuy){
    readyQput(thisThread);
    threadHandoffCPU(nextGuy, Ready);
  }
}


void threadExit()
{
  Thread* nextGuy;

  nextGuy = readyQget();
  assert(nextGuy);
  threadHandoffCPU(nextGuy, Defunct);
}


void threadUsesFloatingPoint(int usesIt)
{
  thisThread->usesFloatingPoint = usesIt;
}


void nodeExit()
{
#ifdef PVM
  leave();
#endif
#ifdef PVM3
  pvm_exit();
#endif
  exit(0);
}


/*****************************************************************************
 *
 *  mutexes
 */

void mutexInit(Mutex* m)
{
  m->locked = 0;
  m->waiters = (Thread*)0;
}


void mutexDestroy(Mutex* m)
{
  assert(m->waiters == (Thread*)0);
}


void mutexAcquire(Mutex* m)
{
  if(m->locked){
#if defined(OLD_MUTEX_BEHAVIOUR)
	  /* "stack" mutex waiters */
	thisThread->next = m->waiters;
	m->waiters = thisThread;
#else
	Thread **t;

	  /* Queue requests for fair access to the mutex */
	thisThread->next = (Thread *)0;

	t = &m->waiters;
	while (*t) 
		t = &(*t)->next;
	*t = thisThread;
#endif
	threadBlock();
  }else{
	m->locked = 1;
  }
}


void mutexRelease(Mutex* m)
{
  Thread* t;

  if(m->waiters){
	t = m->waiters;
	m->waiters = t->next;
	threadUnblock(t);
  }else{
	m->locked = 0;
  }
}


void conditionInit(Condition* c)
{
  c->waiters = (Thread*)0;
}


void conditionDestroy(Condition* c)
{
  assert(c->waiters == (Thread*)0);
}


void conditionWait(Condition* c, Mutex* m)
{
#if defined(OLD_CONDITION_BEHAVIOUR)
  thisThread->next = c->waiters;
  c->waiters = thisThread;
#else
  Thread **t;
  /* Queue requests for fair access to the condition */
  thisThread->next = (Thread *)0;

  t = &c->waiters;
  while (*t)
	  t = &(*t)->next;
  *t = thisThread;
#endif
  mutexRelease(m);
  threadBlock();
  mutexAcquire(m);
}


void conditionSignal(Condition* c)
{
  if(c->waiters){
	threadUnblock(c->waiters);
	c->waiters = c->waiters->next;
  }
}


void conditionBroadcast(Condition* c)
{
  while(c->waiters){
	threadUnblock(c->waiters);
	c->waiters = c->waiters->next;
  }
}


